"""This module handles all API calls"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import str
from requests import HTTPError
from requests import Session
from urllib.parse import (urlencode,
                          urljoin,
                          urlunsplit,
                          urlsplit)

from gobble.config import settings
from gobble.logger import log
from gobble.snapshot import SnapShot, to_json


# One session for all queries
_session = Session()


class EndPoint(object):
    """An API endpoint served by the os-conductor app"""

    def __init__(self, method, *path,
                 trailing_slash=False,
                 session=_session):

        self.has_slash = trailing_slash
        self.method = method
        self._path = path
        self._session = session
        self.snapshot = None

    @property
    def fire(self):
        return getattr(self._session, self.method.lower())

    @property
    def url(self):
        return urljoin(settings.OS_URL, '/'.join(self.path))

    @property
    def path(self):
        """Deal with trailing slashes
        """
        path = list(self._path)
        if self.has_slash:
            path[-1] += '/'
        return path

    def _append(self, params):
        """Encode the query string and append it to the endpoint
        """
        query = urlencode(tuple(params.items()))
        parts = urlsplit(self.url)._replace(query=query)
        return urlunsplit(parts)

    def __call__(self, params=None, **request):
        """Fire, take a snapshot and return the response
        """
        request_url = self._append(params) if params else self.url
        response = self.fire(request_url, **request)
        summary = request_url, response, params
        self.snapshot = SnapShot(self, *summary, **request)
        return response

    def __str__(self):
        return '%s: /%s' % (self.method, '/'.join(self.path))

    def __repr__(self):
        return '<EndPoint ' + str(self) + '>'

    @property
    def info(self):
        return {
            'method': self.method,
            'path': self.path,
            'endslash': self.has_slash,
            "url": self.url
        }


# Expose callable endpoint objects to other modules
# -----------------------------------------------------------------------------
authenticate_user = EndPoint('GET', 'user', 'check')
authorize_user = EndPoint('GET', 'user', 'authorize')
oauth_callback = EndPoint('GET', 'oauth', 'callback')
update_user = EndPoint('POST', 'user', 'update')
search_packages = EndPoint('GET', 'search', 'package')
upload_package = EndPoint('POST', 'package', 'upload')
upload_status = EndPoint('GET', 'package', 'status')
request_upload = EndPoint('POST', 'datastore', trailing_slash=True)
toggle_publish = EndPoint('POST', 'package', 'publish')
# -----------------------------------------------------------------------------


def handle(response):
    """Handle a response and return its payload

    If all is okay, return the json payload of the response
    as a dict. If the status code is in the range 400 to 599,
    raise an HTTPError.
    """
    try:
        response.raise_for_status()
    except HTTPError as error:
        log.error(error)
        raise error

    return to_json(response)



"""Search for users and package inside the ElasticSearch database"""


from gobble.api import search_packages, handle
from gobble.user import User

SEARCH_KEYS = ['q', 'size', 'title', 'author', 'description',
               'regionCode', 'countryCode', 'cityCode']


def search(query, private=True, limit=None):
    """Query the packages on Open-Spending.

    You can search a package by `title`, `author`, `description`, `regionCode`,
    `countryCode` or`cityCode`. You can match all these fields at once with the
    magic `q` key.

    If authentication-token was provided, then private packages from the
    authenticated user will also be included. Otherwise, only public packages
    will be returned. You can limit the size of your results with the `size`
    parameter.

    :param query: a `dict` of key value pairs
    :param private: show private datapackages
    :param limit: the number of results returned

    :type query: `dict`
    :rtype private: `bool'
    :rtype size: `int'

    :return: a dictionary with the results
    :rtype: :class: `list` of `dict`
    """
    _validate(query)

    if limit:
        query.update(size=limit)

    quoted_query = _sanitize(query)

    if private:
        user = User()
        quoted_query.update(jwt=user.token)

    response = search_packages(params=quoted_query)
    return handle(response)


def _sanitize(query):
    keys = ['package.' + str(k) for k in query.keys()]
    values = ['"' + str(v) + '"' for v in query.values()]
    return dict(zip(keys, values))


def _validate(query):
    for key in query.keys():
        if key not in SEARCH_KEYS:
            msg = 'Invalid search key "{key}" for package'
            raise ValueError(msg.format(key=key))



"""Command line interface for Gobble"""

import io

from json import loads
from click import command, argument, Path, group, option, echo
from os.path import join

from gobble.user import create_user
from gobble.fiscal import FiscalDataPackage
from gobble.config import ROOT_DIR
from gobble.search import search


@group()
def gobble():
    pass


@command()
@argument('filepath', type=Path(exists=True))
def version():
    filename = join(ROOT_DIR, 'package.json')
    with io.open(filename) as json:
        package = loads(json.read())
        echo(package['version'])


@command()
def start():
    create_user()


@command()
@argument('filepath',
          type=Path(exists=True),
          metavar='full path to JSON datapackage descriptor')
def check(filepath):
    package = FiscalDataPackage(filepath)
    package.validate(raise_error=False)


@command()
@argument('query_all')
@option('--private', default=True, help='include private packages')
@option('--limit', default=50, help='limit the search results')
@option('--title', default=None, help='search package titles')
@option('--author', default=None, help='search packages by authors')
@option('--description', default=None, help='search packages by descriptions')
@option('--region', default=None, help='search packages by regions')
@option('--country', default=None, help='search packages by 2-digit country code')
@option('--city', default=None, help='search packages by city')
@option('--any', default=None, help='search packages by city')
def pull(private,
         limit,
         query_all,
         title,
         author,
         description,
         region,
         country,
         city):

    search_options = (
        ('title', title),
        ('author', author),
        ('description', description),
        ('regionCode', region),
        ('countryCode', country),
        ('cityCode', city)
    )

    if query_all:
        query = {'q': all}
    else:
        query = {key: value for key, value in search_options if value}

    search(query, private=private, limit=limit)


@command()
@argument('filepath', type=Path(exists=True))
def push(filepath):
    package = FiscalDataPackage(filepath)
    package.upload()


gobble.add_command(version)
gobble.add_command(check)
gobble.add_command(pull)
gobble.add_command(push)
gobble.add_command(start)
